// Original PSG emulator by Maxim Gredy
// Modified by Rodrigo Cardoso for SegaX

#include <string.h>
#include <limits.h>
#include "psg.h"
#include "def.h"

struct _PSG PSG;

// Constants
#define NoiseInitialState  0x8000
#define NoiseWhiteFeedback 0x0009

// These values are taken from a real SMS2's output
const int PSGVolumeValues[2][16] = {
	{ 892, 892, 892, 760, 623, 497, 404, 323, 257, 198, 159, 123, 96, 75, 60, 0 },
	{ 892, 774, 669, 575, 492, 417, 351, 292, 239, 192, 150, 113, 80, 50, 24, 0 }
};

// Value below which PSG does not output
#define PSG_CUTOFF 0x06

void PSG_Init(unsigned long Clock, unsigned long SampleRate)
{
  PSG.dClock = (float)Clock / 16 / SampleRate;
  PSG_Reset();
}

void PSG_Reset(void)
{
  float clock;
  int   i;

  clock = PSG.dClock;
  memset(&PSG, 0, sizeof(PSG));
  PSG.dClock  = clock;
  PSG.Channel = 4;
  PSG.Stereo  = 0xFF;
  for(i = 0; i < 4; i++)
  {
    // Initialise PSG state
    PSG.Registers[2 * i]     = 1;        // tone freq=1
    PSG.Registers[2 * i + 1] = 0x0F;     // vol=off
    PSG.NoiseFreq            = 0x10;
    PSG.ToneFreqPos[i]       = 1;        // Set flip-flops to 1
    PSG.IntermediatePos[i]   = LONG_MIN; // Set intermediate positions to do-not-use value
  }
  PSG.WhiteNoiseFeedback = NoiseWhiteFeedback;
  PSG.NoiseShiftRegister = NoiseInitialState; // Initialise noise generator
  PSG.Mute               = 0x0F;
}

void PSG_Write(unsigned char data)
{
  int latch;

  if(data & 0x80)
  {
	  // Latch/data byte	%1 cc t dddd
    PSG.LatchedRegister  = latch = ((data >> 4) & 0x07);
	  PSG.Registers[latch] = (PSG.Registers[latch] & 0x03F0) | // zero low 4 bits
		                       (data & 0x0F);                    // and replace with data
  } 
  else
  {
    latch = PSG.LatchedRegister;
	  // Data byte		%0 - dddddd
	  if(!(latch & 0x01) && (latch < 5))
    {
		  // Tone register
		  PSG.Registers[latch] = (PSG.Registers[latch] & 0x0F) | // zero high 6 bits
			                       ((data & 0x3F) << 4);           // and replace with data
    }
	  else
    {
		  // Other register
		  PSG.Registers[latch] = data & 0x0F; // Replace with data
    }
  }

  switch(latch)
  {
    case 0:
    case 2:
    case 4:	// Tone channels
      if(PSG.Registers[latch] == 0) PSG.Registers[latch] = 1; // Zero frequency changed to 1 to avoid div/0
      break;
    case 6:	// Noise
      PSG.NoiseShiftRegister = NoiseInitialState;                // reset shift register
      PSG.NoiseFreq          = 0x10 << (PSG.Registers[6] & 0x03); // set noise signal generator frequency
      break;
  }
}

void PSG_Update(int **buf, int len)
{
  int i, j;

  for(j = 0; j < len; buf[0]++, buf[1]++, j++)
  {
    for(i = 0; i < 3; i++)
    {
	    if(PSG.IntermediatePos[i] != LONG_MIN)
		    PSG.Channels[i] = (short)(
          (PSG.Mute >> i & 0x01) * PSGVolumeValues[PSG.VolumeArray][PSG.Registers[2 * i + 1]] * PSG.IntermediatePos[i] / 65536);
	    else
		    PSG.Channels[i] = (
          PSG.Mute >> i & 0x01) * PSGVolumeValues[PSG.VolumeArray][PSG.Registers[2 * i + 1]] * PSG.ToneFreqPos[i];
    }

    PSG.Channels[3] = (short)((PSG.Mute >> 3 & 0x01) * PSGVolumeValues[PSG.VolumeArray][PSG.Registers[7]] * (PSG.NoiseShiftRegister & 0x01));

    *(buf[0]) = *(buf[1]) = 0;
    for(i = 0; i < 4; i++)
    {
      *(buf[0]) += (PSG.Stereo >> (i + 4) & 0x01) * PSG.Channels[i]; // left
      *(buf[1]) += (PSG.Stereo >>  i      & 0x01) * PSG.Channels[i]; // right
    }

    PSG.Clock             += PSG.dClock;
    PSG.NumClocksForSample = (int)PSG.Clock;         // truncates
    PSG.Clock             -= PSG.NumClocksForSample; // remove integer part

    // Decrement tone channel counters
    for(i = 0; i < 3; i++) PSG.ToneFreqVals[i] -= PSG.NumClocksForSample;

    // Noise channel: match to tone2 or decrement its counter
    if(PSG.NoiseFreq == 0x80) PSG.ToneFreqVals[3]  = PSG.ToneFreqVals[2];
    else                     PSG.ToneFreqVals[3] -= PSG.NumClocksForSample;

	  // Tone channels:
	  for(i = 0; i < 3; i++)
    {
		  if(PSG.ToneFreqVals[i] <= 0) // If it gets below 0...
      {
			  if(PSG.Registers[i * 2] > PSG_CUTOFF)
        {
          // Calculate how much of the sample is + and how much is -
          // Go to floating point and include the clock fraction for extreme accuracy :D
          // Store as long int, maybe it's faster? I'm not very good at this
          PSG.IntermediatePos[i] = (long)((PSG.NumClocksForSample - PSG.Clock + 2 * PSG.ToneFreqVals[i]) * PSG.ToneFreqPos[i] / (PSG.NumClocksForSample + PSG.Clock) * 65536);
          PSG.ToneFreqPos[i]     = -PSG.ToneFreqPos[i]; // Flip the flip-flop
			  }
        else
        {
          PSG.ToneFreqPos[i]     = 1; // stuck value
          PSG.IntermediatePos[i] = LONG_MIN;
			  }
			  PSG.ToneFreqVals[i] += PSG.Registers[i * 2] * (PSG.NumClocksForSample / PSG.Registers[i * 2] + 1);
		  }
      else PSG.IntermediatePos[i] = LONG_MIN;
	  }

/*
The player works by calculating the number of PSG clocks per sample (dClock).
For each sample, it renders an integer number of PSG clocks (NumClocksForSample)
and any fraction is saved (Clock) and added into the next sample's number of
clocks. This "saved" fraction is just making sure that the PSG clock flips the
right number of times since for some samples it has to flip more times to stay
at the right speed. The rendered sample still reflects the whole period.

Number of PSG clocks for this sample = dClock (float)
Integer part of that = NumClocksForSample
Fractional part of that = Clock

ToneFreqVals[i] = underflowed sample counter for channel i, ie. a negative integer

|   |   |   |   |   |   |   |   | PSG clocks
  |                            |  Sample boundaries
  .                            .
----------------+              .
  .             |              .  Waveform
  .             |              .
  .             +----------------
  .             .              .
  |   |   |   | . |   |   |   ||  Re-aligned PSG clocks
  <------ A ---->             ..
  .           . <----- B ------>
  .           .               ..
  <---- NumClocksForSample --->.
  .           .               >< Clock
              <------ D ------>

If the first half-wave has sign S, the average value over the sample is:
   A*S+B*(-S)   (A-B)*S
   ---------- = -------
      A+B         A+B

A+B = NumClocksForSample+Clock (integer + fraction of aligned PSG clocks)

The half-wave counter will have just overflowed so that

D = -ToneFreqVals[i]

A = NumClocksForSample-D
  = NumClocksForSample+ToneFreqVals[i]
B = NumClocksForSample+Clock-A
  = NumClocksForSample+Clock-(NumClocksForSample+ToneFreqVals[i])
  = Clock-ToneFreqVals[i]
S = ToneFreqPos[i]

Thus,
                (NumClocksForSample+ToneFreqVals[i]) - (Clock-ToneFreqVals[i]) * ToneFreqPos[i]
Average value = -------------------------------------------------------------------------------
                                           NumClocksForSample+Clock

              = ((NumClocksForSample-Clock+2*ToneFreqVals[i])*ToneFreqPos[i])/(float)(NumClocksForSample+Clock)

However, as a small and probably useless optimisation, I try to avoid storing
this in floating point by using fixed point; I do the calculation (I expect it
is performed in double precision), multiply the result by 64K and round to an
integer. When using this "intermediate position" value, the resulting sample
is divided by 64K to get back to the correct value (ie. I avoid a floating-
*/

	  // Noise channel
	  if(PSG.ToneFreqVals[3] <= 0) // If it gets below 0...
    {
		  PSG.ToneFreqPos[3] *= -1; // Flip the flip-flop
      // If not matching tone2, decrement counter
		  if(PSG.NoiseFreq != 0x80) PSG.ToneFreqVals[3] += PSG.NoiseFreq * (PSG.NumClocksForSample / PSG.NoiseFreq + 1);
		  if(PSG.ToneFreqPos[3] == 1) // Only once per cycle...
      {
			  int Feedback, n;
			  if(PSG.Registers[6] & 0x04)
        {
          // White noise
				  // Calculate parity of fed-back bits for feedback
				  switch(PSG.WhiteNoiseFeedback)
          {
            // Do some optimised calculations for common (known) feedback values
            case 0x0006:	// SC-3000		%00000110
            case 0x0009:	// SMS, GG, MD	%00001001
              n = PSG.NoiseShiftRegister & PSG.WhiteNoiseFeedback;
              // If two bits fed back, I can do Feedback=(nsr & fb) && (nsr & fb ^ fb)
              // since that's (one or more bits set) && (not all bits set)
              Feedback = n && (n ^ PSG.WhiteNoiseFeedback);
              break;
            case 0x8005:	// BBC Micro
              // fall through :P can't be bothered to think too much
            default:		// Default handler for all other feedback values
	            Feedback  = PSG.NoiseShiftRegister & PSG.WhiteNoiseFeedback;
	            Feedback ^= Feedback >> 8;
	            Feedback ^= Feedback >> 4;
	            Feedback ^= Feedback >> 2;
	            Feedback ^= Feedback >> 1;
	            Feedback &= 1;
	            break;
				  }
			  }
        else Feedback = PSG.NoiseShiftRegister & 1; // Periodic noise

			  PSG.NoiseShiftRegister = (PSG.NoiseShiftRegister >> 1) | (Feedback << 15);

/*
// Original code:
      PSG.NoiseShiftRegister = (PSG.NoiseShiftRegister >> 1) |
        ((PSG.Registers[6] & 0x04 ?
        ((PSG.NoiseShiftRegister & 0x9) && (PSG.NoiseShiftRegister & 0x09 ^ 0x09)) :
         PSG.NoiseShiftRegister & 1) << 15);
*/
		  }
	  }
  }
}
