/* script.c - this file is part of segax project. Rodrigo Cardoso.
** Script file menagement. */

#include <string.h>
#include <malloc.h>
#include "script.h"
#include "util.h"

// use temporary file to store strings
#ifdef SCRIPT_USE_TEMP_FILE

// temporary file last access
typedef enum _STMPLA {
  _NONE_,
  _READ_,
  _WRITE_
} STMPLA;

// temporary file data
typedef struct _STMPFILE {
  FILE  *file;
  long   cur_pos;
  int    is_eof;
  STMPLA last_access;
  char   path[MAX_PATH];
} STMPFILE;

STMPFILE *script_tmp_create(void)
{
  STMPFILE *f;
  f = (STMPFILE *)malloc(sizeof(STMPFILE));
  if(!f) return (STMPFILE *)0;
  GetTempFilePath(f->path);
  f->file = fopen(f->path, "w+b");
  if(!f->file) { free(f); return (STMPFILE *)0; }
  f->cur_pos     = 0;
  f->is_eof      = 1;
  f->last_access = _NONE_;
  return f;
}

void script_tmp_close(STMPFILE *f)
{
  if(f)
  {
    if(f->file) { fclose(f->file); remove(f->path); }
    free(f);
  }
}

long script_tmp_write(STMPFILE *f, long pos, void *ptr, long size)
{
  long l;
  if(!f || !ptr) return -1;
  if(pos < 0)
  {
    if(f->last_access == _READ_ || !f->is_eof)
    {
      if(fseek(f->file, 0, SEEK_END)) return -1;
      f->cur_pos     = ftell(f->file);
      f->is_eof      = 1;
      f->last_access = _NONE_;
    }
  }
  else
  {
    if(f->last_access == _READ_ || f->cur_pos != pos)
    {
      if(fseek(f->file, pos, SEEK_SET)) return -1;
      f->cur_pos     = pos;
      f->is_eof      = 0;
      f->last_access = _NONE_;
    }
  }

  if(fwrite(ptr, size, 1, f->file) != 1) return -1;
  l              = f->cur_pos;
  f->last_access = _WRITE_;
  f->cur_pos    += size;
  if(!f->is_eof) f->is_eof = feof(f->file);
  return l;
}

long script_tmp_read(STMPFILE *f, long pos, void *ptr, long size)
{
  long l;
  if(!f || !ptr || pos < 0) return -1;
  if(f->last_access == _WRITE_ || f->cur_pos != pos)
  {
    if(fseek(f->file, pos, SEEK_SET)) return -1;
    f->cur_pos     = pos;
    f->last_access = _NONE_;
  }
  if(fread(ptr, size, 1, f->file) != 1) return -1;
  l              = f->cur_pos;
  f->last_access = _READ_;
  f->cur_pos    += size;
  f->is_eof      = 0;
  return l;
}

#endif

#define INVALID_NAME_CHAR(c) ((c) == '#' || (c) == '=' || (c) == ';' || (c) == '\"' || (c) == '\'' || (c) == ',' || (c) == '{' || (c) == '}')
#define DEF_NAME_CHAR '_'

typedef enum _SCRIPT_READ_TYPE {
    _OTHER_,           // a number or a string not enclosed by "" or ''
    _NULL_,            // there is no data, just a name
    _STRING_,          // data is enclosed by "" or ''
    _MACRO_,           // it's a macro, begining with #
    _NULL_MACRO_,      // it's an empity macro
    _MULTI_            // data is multiple numbers enclosed by { }
} SCRIPT_READ_TYPE;

// used to read script files 
typedef struct _SCRIPT_READ_STRUCT {
  FILE            *file;
  char            *data_name;
  char            *data;
  int              data_size;
  int              data_name_mem_size;
  int              data_mem_size;
  FILE            *ftmp;          // temporary file to store multiple integers
  char             tfn[MAX_PATH]; // temporary file path
  int              tmp_items;     // number of integers in temporary file
  SCRIPT_READ_TYPE type;
} SCRIPT_READ_STRUCT;

const char      script_null_str[]  = "";
int             script_escape_ch[] = { '\\', '\"', '\'', 'a',  'b',  'n',  'f',  'r',  't',  'v',  '?', '0', -1 };
int             script_escape_eq[] = { '\\', '\"', '\'', '\a', '\b', '\n', '\f', '\r', '\t', '\v', '\?', 0,  -1 };

SCRIPT_READ_STRUCT *script_read_init(const char *file_name)
{
  FILE               *file;
  SCRIPT_READ_STRUCT *s;

  file = fopen(file_name, "rt");
  if(!file) return (SCRIPT_READ_STRUCT *)0;

  s = (SCRIPT_READ_STRUCT *)malloc(sizeof(SCRIPT_READ_STRUCT));
  if(!s) { fclose(file); return (SCRIPT_READ_STRUCT *)0; }

  s->file               = file;
  s->data_name_mem_size = s->data_mem_size = 1024;
  s->data_name          = (char *)malloc(s->data_name_mem_size);
  if(!(s->data_name)) { fclose(file); free(s); return (SCRIPT_READ_STRUCT *)0; }
  s->data = (char *)malloc(s->data_mem_size);
  if(!(s->data))
  { fclose(file); free(s->data_name); free(s); return (SCRIPT_READ_STRUCT *)0; }
  s->ftmp = (FILE *)0;

  return s;
}

void script_read_quit(SCRIPT_READ_STRUCT *s)
{
  if(s->ftmp) { fclose(s->ftmp); remove(s->tfn); }
  fclose(s->file);
  free(s->data);
  free(s->data_name);
  free(s);
}

#define WRITE_CHAR(ch) {                          \
  if     (++i >= 1024 * 16) return 0;             \
  else if(*buf_size <= i)                         \
  {                                               \
    p = (char *)realloc(*buf, *buf_size + 1024);  \
    if(!p) return 0;                              \
    *buf       = p;                               \
    *buf_size += 1024;                            \
  }                                               \
  (*buf)[i - 1] = (char)ch; }


int script_read_next_data(SCRIPT_READ_STRUCT *s)
{
  FILE *f;
  char *p, **buf;
  int   c, i, n, ec, spaces, *buf_size, string_char;
  enum {
    SEEKING_NAME,
    DIVIDE_CHAR,
    READING_NAME,
    SEEKING_DATA,
    READING_DATA,
    READING_STRING,
    FINISHED_STRING,
    ESCAPE_CHAR,
    COMMENT,
    LINE_COMMENT,
    MULTIPLY_CHAR,
    SEEKING_MACRO_NAME,
    READING_MACRO_NAME,
    SEEKING_MACRO_DATA,
    READING_MACRO_DATA,
    READING_MACRO_STRING,
    READING_MULTI,
  } state, last_state;

  f = s->file;
  if(feof(f)) return 0;

  *(s->data_name) = *(s->data) = 0;
  s->data_size = 0;
  s->type      = _OTHER_;

  // reading next data of the file
  state    = last_state = SEEKING_NAME;
  i        = spaces = string_char = 0;
  buf      = &(s->data_name);
  buf_size = &(s->data_name_mem_size);
  while((c = getc(f)) != EOF)
  {
    switch(state)
    {
      case SEEKING_NAME:                                              // searching for data name
        if     (ISSPACE(c) || c == ';') break;                        // skiping spaces before name
        else if(c == '=')               return 0;                     // error: no name!
        else if(c == '/')             { state = DIVIDE_CHAR; break; } // maybe a comment
        else if(c == '#')             { state = last_state = SEEKING_MACRO_NAME; s->type = _MACRO_; break; }
        else                            state = last_state = READING_NAME;

      case READING_NAME:
        if     (c == '/')                           state = DIVIDE_CHAR; // maybe a comment, wait for next character
        else if(c == '\"' || c == '\'' || c == '#') return 0;            // invalid character
        else if(c == '='  || c == ';')                                   // name end
        {
          WRITE_CHAR(0);
          i        = 0;
          buf      = &(s->data);
          buf_size = &(s->data_mem_size);
          if(c == ';') { s->type = _NULL_; goto end; }
          else           state = last_state = SEEKING_DATA;
        }
        else if(!ISSPACE(c))                                             // remove spaces from the name
        {
          WRITE_CHAR(c);
        }
        break;

      case SEEKING_DATA:                                  // searching for data
        if     (ISSPACE(c)) break;                        // skiping spaces before data
        else if(c == '/') { state = DIVIDE_CHAR; break; }
        else if(c == '\"' || c == '\'')                   // it's a string closed by "" or ''
        {
          s->type = _STRING_;                             // data type = string
          state = last_state = READING_STRING;
          string_char        = c;
          break;
        }
        else if(c == ';') { s->type = _NULL_; goto end; } // null data
        else if(c == '{')
        {
          s->type = _MULTI_;
          state   = last_state = READING_MULTI;
          if(!s->ftmp)
          {
            GetTempFilePath(s->tfn);
            s->ftmp = fopen(s->tfn, "w+b");
            if(!s->ftmp) return 0;
          }
          else if(fseek(s->ftmp, 0, SEEK_SET)) return 0;
          s->tmp_items = 0;
          break;
        }
        else if(c == '#')   return 0;
        else                state = last_state = READING_DATA;

      case READING_DATA:
        if     (c == '/')  state = DIVIDE_CHAR;     // may be a comment, wait for next character
        else if(c == '\"' || c == '\'') return 0;   // invalid char
        else if(c == ';') goto end;                 // data end
        else if(c == '='  || c == '#')  return 0;
        else if(c == '{'  || c == '}')  return 0;
        else
        {
          if(ISSPACE(c)) spaces++;
          else           spaces = 0;
          WRITE_CHAR(c);
        }
        break;

      case READING_STRING:
        if     (c == string_char) state = last_state = FINISHED_STRING; // string end
        else if(c == '\\')        state = ESCAPE_CHAR;                  // can be a string escape character
        else if(c == '\n')        return 0;
        else                      WRITE_CHAR(c);
        break;

      case FINISHED_STRING:
        if     (c == ';')         goto end;            // data end
        else if(c == string_char) state = last_state = READING_STRING;
        else if(c == '/')         state = DIVIDE_CHAR; // may be a comment
        else if(!ISSPACE(c))      return 0;            // invalid character after string
        break;

      case ESCAPE_CHAR:
        ec = n = 0;
        while(script_escape_ch[n] != -1)
        {
          if(script_escape_ch[n] == c) { ec = 1; break; }
          n++;
        }
        if(ec) // it's a valid escape character
        {
          WRITE_CHAR(script_escape_eq[n]);
        }
        else
        {
          WRITE_CHAR('\\');
          ungetc(c, f);
        }
        state = last_state;
        break;

      case DIVIDE_CHAR:
        if     (c == '*') state = COMMENT;            // comment start
        else if(c == '/') state = LINE_COMMENT;       // line comment start
        else
        {
          switch(last_state)
          {
            case FINISHED_STRING:
              return 0;
            case SEEKING_NAME:
              state = last_state = READING_NAME;
              break;
            case SEEKING_DATA:
              state = last_state = READING_DATA;
              break;
            case SEEKING_MACRO_NAME:
            case READING_MACRO_NAME:
            case SEEKING_MACRO_DATA:
            case READING_MACRO_DATA:
            case READING_MULTI:
              return 0;
            default:
              state = last_state;
          }
          WRITE_CHAR('/');
          ungetc(c, f);
        }
        break;

      case LINE_COMMENT:
        if(c == '\n')
        {
          switch(last_state)
          {
            case SEEKING_MACRO_NAME:
              return 0;
            case SEEKING_MACRO_DATA:
              s->type = _NULL_MACRO_;
            case READING_MACRO_NAME:
            case READING_MACRO_DATA:
              goto end;
            default:
              break;
          }
          state = last_state; // end of comment
        }
        break;

      case COMMENT:
        if(c == '*') state = MULTIPLY_CHAR;
        break;

      case MULTIPLY_CHAR:
        if     (c == '/')
        {
          switch(last_state)
          {
            case SEEKING_MACRO_NAME:
              return 0;
            case SEEKING_MACRO_DATA:
              s->type = _NULL_MACRO_;
            case READING_MACRO_NAME:
            case READING_MACRO_DATA:
              goto end;
            default:
              break;
          }
          state = last_state; // end of comment
        }
        else if(c != '*') state = COMMENT;
        break;

      case SEEKING_MACRO_NAME:
        if     (INVALID_NAME_CHAR(c))  return 0;
        else if(c == '*' || c == '\\') return 0;
        else if(c == '/')              state = DIVIDE_CHAR;
        else if(!ISSPACE(c))         { WRITE_CHAR(c); state = last_state = READING_MACRO_NAME; }
        break;

      case READING_MACRO_NAME:
        if     (ISSPACE(c) || c == '\"' || c == '\'')
        {
          WRITE_CHAR(0);
          state    = last_state = ISSPACE(c) ? SEEKING_MACRO_DATA : READING_MACRO_STRING;
          i        = 0;
          buf      = &(s->data);
          buf_size = &(s->data_mem_size);
        }
        else if(c == '\\' || c == '*')  return 0;
        else if(c == '/')               state = DIVIDE_CHAR;
        else if(c == '\n')              goto end;
        else if(INVALID_NAME_CHAR(c))   return 0;
        else                            WRITE_CHAR(c);
        break;

      case SEEKING_MACRO_DATA:
        if     (c == '\n')                                      goto end;
        if     (c == '#'  || c == ';' || c == '=' || c == '\\') return 0;
        else if(c == '\\' || c == '*')                          return 0;
        else if(c == '/')                                       state = DIVIDE_CHAR;
        else if(c == '\"' || c == '\'')                         state = last_state = READING_MACRO_STRING;
        else if(!ISSPACE(c))
        {
          WRITE_CHAR(c);
          state = last_state = READING_MACRO_DATA;
        }
        break;

      case READING_MACRO_DATA:
        if     (c == '\n')              goto end;
        else if(c == '/')               state = DIVIDE_CHAR;
        else if(c == '\"' || c == '\'') state = last_state = READING_MACRO_STRING;
        else if(c == '\\' || c == '#')  return 0;
        else if(c == '=' || c == '*')   return 0;
        else                            WRITE_CHAR(c);
        break;

      case READING_MACRO_STRING:
        if     (c == '\n')              return 0;
        else if(c == '\"' || c == '\'') state = last_state = SEEKING_MACRO_DATA;
        else if(c == '/')               state = ESCAPE_CHAR;
        else                            WRITE_CHAR(c);
        break;

      case READING_MULTI:
        if(!ISSPACE(c))
        {
          if     (c == '}' || c == ',')
          {
            WRITE_CHAR(0);
            n = Str2Int(s->data, 0, &i);
            if(i) { i = 0; return 0; } // invalid number
            i = 0;
            if(fwrite(&n, 4, 1, s->ftmp) != 1) return 0;
            s->tmp_items++;
            if(c == '}') goto end;
          }
          else if(c == '/') state = DIVIDE_CHAR;
          else if(ISHEXDIGIT(c) || c == 'x' || c == 'X')
          {
            WRITE_CHAR(c);
          }
          else return 0;
        }
        break;
    }
  }

  if(*s->data_name) return 0;
  end:
  i -= spaces;
  WRITE_CHAR(0);
  s->data_size = i;

  return 1;
}

// compare two names, return TRUE if equal
BOOL script_compare_names(const char *n1, const char *n2)
{
  int c1, c2, s;

  c1 = (int)*n1;
  c2 = (int)*n2;
  while(c1 || c2)
  {
    s = 1;
    if(ISSPACE(c1)) { c1 = (int)*n1++; s = 0; }
    if(ISSPACE(c2)) { c2 = (int)*n2++; s = 0; }
    if(s)
    {
      // converting characters to uppercase
      if(c1 >= 'a' && c1 <= 'z')     c1 -= ('a' - 'A');
      else if(INVALID_NAME_CHAR(c1)) c1 = DEF_NAME_CHAR;
      if(c2 >= 'a' && c2 <= 'z') c2 -= ('a' - 'A');
      else if(INVALID_NAME_CHAR(c2)) c2 = DEF_NAME_CHAR;
      if(c1 != c2) return FALSE;
      if(c1) c1 = (int)*n1++;
      if(c2) c2 = (int)*n2++;
    }
  }

  return TRUE;
}

// add a null data to a script
BOOL script_write_null_data(SCRIPT_FILE *f, char *name)
{
  char        *ptr, *s;
  SCRIPT_DATA *p;
  int          i, l, max;

  for(max = f->n_data, i = 0; i < max; i++)
  { if(script_compare_names(f->data[i].name, name)) return 1; }

  if(f->n_alloc <= f->n_data) // must expand memory
  {
    p = (SCRIPT_DATA *)realloc(f->data, (f->n_alloc + 32) * sizeof(SCRIPT_DATA));
    if(!p) return FALSE;
    f->data    = p;
    f->n_alloc += 32;
  }

  p = f->data + f->n_data++;

  // get name size
  l = 1;
  s = (char *)name;
  while((i = *(s++))) { if(!ISSPACE(i)) l++; }

  // alloc memory and copy name
  if(f->base_size < f->names_size + l)
  {
    ptr = f->names_base ? (char *)realloc(f->names_base, f->base_size + l + 1024) :
                          (char *)malloc(f->base_size + l + 1024);
    if(!ptr) return FALSE;
    f->names_base    = ptr;
    f->base_size    += l + 1024;
    f->data[0].name = ptr;
    for(i = 1; i < f->n_data; i++)
      f->data[i].name = f->data[i - 1].name + f->data[i - 1].name_size;
  }
  else p->name = f->names_base + f->names_size;
  f->names_size += l;
  p->name_size   = l;
  ptr            = p->name;
  s              = (char *)name;
  while((i = *(s++)))
  {
    if(!ISSPACE(i))
    {
      if(INVALID_NAME_CHAR(i)) i = DEF_NAME_CHAR;
      *ptr++ = (char)i;
    }
  }
  *ptr = 0;

  p->n_data    = 0;
  p->data_size = 4;
  p->data_type = SDT_NULL;

  return TRUE;
}

BOOL script_process_macro(SCRIPT_FILE *, const char *, const char *, const char *);

// load script file
BOOL script_load_file(const char *path, SCRIPT_FILE *f)
{
  SCRIPT_READ_STRUCT *s;
  int                 n, error;

  s = script_read_init(path);
  if(!s) return FALSE;

  while(script_read_next_data(s))
  {
    if(*s->data_name)
    {
      if     (s->type == _MACRO_)
      {
        script_process_macro(f, path, s->data_name, s->data); // macro
      }
      else if(s->type == _NULL_)
      {
        if(!script_write_null_data(f, s->data_name)) // null data
        { Script_Free(f); f = (SCRIPT_FILE *)0; break; }
      }
      else if(s->type == _MULTI_) // multiple numbers
      {
        char *ps;
        int   ns = s->tmp_items * 4;
        // checking if data buffer is enought to store converted data
        if(s->data_mem_size < ns)
        {
          ps = (char *)realloc(s->data, ns);
          if(!ps) { Script_Free(f); f = (SCRIPT_FILE *)0; break; }
          s->data          = ps;
          s->data_mem_size = ns;
        }
        // reading numbers
        if(fseek(s->ftmp, 0, SEEK_SET))         { Script_Free(f); f = (SCRIPT_FILE *)0; break; }
        if(fread(s->data, ns, 1, s->ftmp) != 1) { Script_Free(f); f = (SCRIPT_FILE *)0; break; }
        if(!Script_WriteMultiInts(f, s->data_name, (const int *)s->data, s->tmp_items))
        { Script_Free(f); f = (SCRIPT_FILE *)0; break; }
      }
      else
      {
        if(s->type == _OTHER_)                      // may be a number or BOOL
        {
          if(!strcmpi(s->data, "FALSE"))            // BOOL
          {
            if(!Script_WriteBool(f, s->data_name, FALSE))
            { Script_Free(f); f = (SCRIPT_FILE *)0; break; }
          }
          else if(!strcmpi(s->data, "TRUE"))        // BOOL
          {
            if(!Script_WriteBool(f, s->data_name, TRUE))
            { Script_Free(f); f = (SCRIPT_FILE *)0; break; }
          }
          else
          {
            n = Str2Int(s->data, 0, &error);        // number
            if     (error) s->type = _STRING_;      // invalid number
            else if(!Script_WriteInt(f, s->data_name, n))
            { Script_Free(f); f = (SCRIPT_FILE *)0; break; }
          }
        }

        if(s->type == _STRING_)
        {
          if(!Script_WriteString(f, s->data_name, s->data, s->data_size))
          { Script_Free(f); f = (SCRIPT_FILE *)0; break; }
        }
      }
    }
  }

  script_read_quit(s);

  return TRUE;
}

BOOL script_process_macro(SCRIPT_FILE *f, const char *path, const char *name, const char *data)
{
  if(!strcmpi(name, "include")) // include macro
  {
    int   i, c, lc, is_full_path;
    char *p;
    BOOL  b;

    while(ISSPACE(*data)) data++;
    switch(*data)
    {
      case '\"': lc = '\"'; data++; break;
      case '\'': lc = '\''; data++; break;
      case '<':  lc = '>';  data++; break;
      default:   lc = 0;    break;
    }
    i = is_full_path = 0;
    p = (char *)data;
    while((c = *(p++)))
    {
      if     (c == lc)               break;
      else if(c == '\\' || c == '/') is_full_path = 1;
      i++;
    }
    if(c != lc) return FALSE;
    p -= 2;
    while(ISSPACE(*p--)) i--;
    if(is_full_path)
    {
      p = (char *)malloc(i + 1);
      if(!p) return FALSE;
      memcpy(p, data, i);
      p[i] = 0;
    }
    else
    {
      c = strlen(path);
      p = (char *)path + c - 1;
      while(c > 0 && *p != '\\' && *p != '/') { c--; p--; }
      p = (char *)malloc(i + c + 1);
      if(!p) return FALSE;
      memcpy(p, path, c);
      memcpy(p + c, data, i);
      p[c + i] = 0;
    }
    b = script_load_file(p, f);
    free(p);
    return b;
  }
  return FALSE;
}

// create a new script file in memory
SCRIPT_FILE *Script_Create(void)
{
  SCRIPT_FILE *f;

  f = (SCRIPT_FILE *)malloc(sizeof(SCRIPT_FILE));
  if(!f) return (SCRIPT_FILE *)0;
  f->n_data   = 0;
  f->n_alloc    = 32;
  f->names_base = (char *)0;
  f->names_size = f->base_size = 0;
  f->data      = (SCRIPT_DATA *)malloc(f->n_alloc * sizeof(SCRIPT_DATA));
  if(!f) { free(f); return (SCRIPT_FILE *)0; }

#ifdef SCRIPT_USE_TEMP_FILE
  f->tmp_file = script_tmp_create();
  if(!f->tmp_file) { free(f->data); free(f); return (SCRIPT_FILE *)0; }
#endif

  return (SCRIPT_FILE *)f;
}

// free the memory used to store the script file
void Script_Free(SCRIPT_FILE *pFile)
{
  if(pFile)
  {
#ifdef SCRIPT_USE_TEMP_FILE
    if(pFile->tmp_file) script_tmp_close((STMPFILE *)pFile->tmp_file);
#else
    int i, max;
    for(max = pFile->n_data, i = 0; i < max; i++)
    {
      if(pFile->data[i].data_type == SDT_STRING) free(pFile->data[i].sz_data);
    }
#endif
    if(pFile->names_base) free(pFile->names_base);
    if(pFile->data)      free(pFile->data);
    free(pFile);
  }
}

// load the script file into memory
SCRIPT_FILE *Script_Load(const char *szFileName)
{
  SCRIPT_FILE *f = Script_Create();
  if(!f) return (SCRIPT_FILE *)0;
  if(!script_load_file(szFileName, f)) return Script_Free(f), (SCRIPT_FILE *)0;
  return f;
}

// save the script on memory into a file
BOOL Script_Save(SCRIPT_FILE *pFile, const char *szName, const char *szComment)
{
  FILE    *f;
  int      i, max;
  SCRIPT_DATA *p;
#ifdef SCRIPT_USE_TEMP_FILE
  char    *buf      = (char *)0;
  int      buf_size = 0;
#endif

  f = fopen(szName, "wt");
  if(!f) return FALSE;

  if(szComment) fprintf(f, "/* %s */\n\n", szComment);

  for(max = pFile->n_data, i = 0; i < max; i++)
  {
    p = pFile->data + i;
    switch(p->data_type)
    {
      case SDT_NULL:
        fprintf(f, "%s = ;\n", p->name);
        break;
      case SDT_NUMBER:
        Script_WriteIntDirect(f, p->name, p->n_data, (const char *)0, 0);
        break;
      case SDT_BOOL:
        Script_WriteBoolDirect(f, p->name, p->n_data, (const char *)0, 0);
        break;
      case SDT_STRING:
#ifdef SCRIPT_USE_TEMP_FILE
        if(buf_size < p->data_size)
        {
          if(buf) free(buf);
          buf = (char *)malloc(p->data_size);
          if(!buf) { fclose(f); return FALSE; }
          buf_size = p->data_size;
        }
        if(script_tmp_read((STMPFILE *)pFile->tmp_file, p->data_pos, buf, p->data_size) == -1)
        { if(buf) free(buf); fclose(f); return FALSE; }
        Script_WriteStringDirect(f, p->name, buf, p->data_size, (const char *)0, 0);
#else
        Script_WriteStringDirect(f, p->name, p->sz_data, p->data_size, (const char *)0, 0);
#endif
        break;
      case SDT_MULTI:
#ifdef SCRIPT_USE_TEMP_FILE
        if(buf_size < p->data_size)
        {
          if(buf) free(buf);
          buf = (char *)malloc(p->data_size);
          if(!buf) { fclose(f); return FALSE; }
          buf_size = p->data_size;
        }
        if(script_tmp_read((STMPFILE *)pFile->tmp_file, p->data_pos, buf, p->data_size) == -1)
        { if(buf) free(buf); fclose(f); return FALSE; }
        Script_WriteMultiIntsDirect(f, p->name, (const int *)buf, p->data_size >> 2, (const char *)0, 0);
#else
        Script_WriteMultiIntsDirect(f, p->name, p->pn_data, p->data_size >> 2, (const char *)0, 0);
#endif
        break;
    }
  }

#ifdef SCRIPT_USE_TEMP_FILE
  if(buf) free(buf);
#endif
  fclose(f);

  return TRUE;
}

// load the specified string in script file
int Script_GetString(SCRIPT_FILE *pFile, const char *szName, const char *szDefault,
                     int nDefSize, char *szBuf, int nBufSize, BOOL *pExist)
{
  SCRIPT_DATA *p;
  int          i, max, l;
  const char  *s;

  if(!szName) return 0;

#ifdef SCRIPT_USE_TEMP_FILE
  if(pFile)
  {
    for(max = pFile->n_data, i = 0; i < max; i++)
    {
      p = pFile->data + i;
      if(script_compare_names(p->name, szName))
      {
        if(p->data_type == SDT_NULL)
        {
          if(szBuf && nBufSize > 0) *szBuf  = 0;
          if(pExist)                *pExist = TRUE;
          return 1;
        }
        else if(p->data_type == SDT_STRING)
        {
          if(szBuf)
          {
            if(nBufSize >= p->data_size)
            { if(script_tmp_read((STMPFILE *)pFile->tmp_file, p->data_pos, szBuf, p->data_size) == -1) goto _default; }
          }
          if(pExist) *pExist = TRUE;
          return p->data_size;
        }
      }
    }
  }

_default:
#else
  s = (const char *)0;

  if(pFile)
  {
    for(max = pFile->n_data, i = 0; i < max; i++)
    {
      p = pFile->data + i;
      if(script_compare_names(p->name, szName)) // data found
      {
        if(p->data_type == SDT_NULL) // this is a null data
        {
          s = script_null_str;
          l = 1;
        }
        else if(p->data_type == SDT_STRING)
        {
          s = p->sz_data;
          l = p->data_size;
        }
        break;
      }
    }
  }

  if(!s)
  {
#endif
    if(szDefault)
    {
      if(nDefSize <= 0) nDefSize = strlen(szDefault) + 1;
      s = szDefault;
      l = nDefSize;
    }
    else { s = script_null_str; l = 1; }

    if(pExist) *pExist = FALSE;
#ifndef SCRIPT_USE_TEMP_FILE
  }
  else if(pExist) *pExist = TRUE;
#endif

  if(szBuf && nBufSize >= l) memcpy(szBuf, s, l);

  return l;
}

// load the specified number in script file
int Script_GetInt(SCRIPT_FILE *pFile, const char *szName, int nDefault, BOOL *pExist)
{
  SCRIPT_DATA *p;
  int          i, max;

  if(!pFile || !szName)
  {
    if(pExist) *pExist = FALSE;
    return nDefault;
  }

  for(max = pFile->n_data, i = 0; i < max; i++)
  {
    p = pFile->data + i;
    if(script_compare_names(p->name, szName)) // data found
    {
      if(p->data_type == SDT_NULL) // this is a null data
      {
        if(pExist) *pExist = TRUE;
        return 0;
      }
      else if(p->data_type == SDT_NUMBER)
      {
        if(pExist) *pExist = TRUE;
        return p->n_data;
      }
      else if(p->data_type == SDT_MULTI)
      {
        if(pExist) *pExist = TRUE;
#ifdef SCRIPT_USE_TEMP_FILE
        if(script_tmp_read((STMPFILE *)pFile->tmp_file, p->data_pos, &i, 4) == -1) return nDefault;
        return i;
#else
        return p->pn_data[0];
#endif
      }
      break;
    }
  }

  if(pExist) *pExist = FALSE;

  return nDefault;
}

// load a BOOL variable
BOOL Script_GetBool(SCRIPT_FILE *pFile, const char *szName, BOOL fDefault, BOOL *pExist)
{
  SCRIPT_DATA *p;
  int          i, max;

  if(!pFile || !szName)
  {
    if(pExist) *pExist = FALSE;
    return fDefault ? TRUE : FALSE;
  }

  for(max = pFile->n_data, i = 0; i < max; i++)
  {
    p = pFile->data + i;
    if(script_compare_names(p->name, szName)) // data found
    {
      if(p->data_type == SDT_NULL) // this is a null data
      {
        if(pExist) *pExist = TRUE;
        return FALSE;
      }
      else if(p->data_type == SDT_NUMBER)
      {
        if(pExist) *pExist = TRUE;
        return p->n_data ? TRUE : FALSE;
      }
      else if(p->data_type == SDT_BOOL)
      {
        if(pExist) *pExist = TRUE;
        return p->n_data;
      }
      break;
    }
  }

  if(pExist) *pExist = FALSE;

  return fDefault ? TRUE : FALSE;
}

int Script_GetMultiInts(SCRIPT_FILE *pFile, const char *szName, const int *pnDefault,
                        int *pnBuf, int nData, BOOL *pExist)
{
  SCRIPT_DATA *p;
  int          i, max, l;
  int          nBufSize = nData * 4;

  if(!szName) return 0;

  if(pFile)
  {
    for(max = pFile->n_data, i = 0; i < max; i++)
    {
      p = pFile->data + i;
      if(script_compare_names(p->name, szName))
      {
        if(p->data_type == SDT_NULL)
        {
          if(pnBuf) { for(i = 0; i < nData; i++) pnBuf[i] = 0; }
          if(pExist) *pExist = TRUE;
          return nData;
        }
        else if(p->data_type == SDT_NUMBER)
        {
          if(pnBuf && nData >= 1)
          {
            *pnBuf = p->n_data;
            for(i = 1; i < nData; i++) pnBuf[i] = 0;
          }
          return 1;
        }
        else if(p->data_type == SDT_MULTI)
        {
          if(pnBuf)
          {
            if(nBufSize >= p->data_size) l = p->data_size;
            else                         l = nBufSize;
#ifdef SCRIPT_USE_TEMP_FILE
            if(script_tmp_read((STMPFILE *)pFile->tmp_file, p->data_pos, pnBuf, l) == -1) goto _default;
#else
            memcpy(pnBuf, p->pn_data, l);
#endif
            for(i = l >> 2; l < nData; l++) pnBuf[i] = 0;
          }
          if(pExist) *pExist = TRUE;
          return p->data_size >> 2;
        }
      }
    }
  }

#ifdef SCRIPT_USE_TEMP_FILE
_default:
#endif
  if(pnBuf)
  {
    if(pnDefault) memcpy(pnBuf, pnDefault, nBufSize);
    else          memset(pnBuf, 0, nBufSize);
  }

  if(pExist) *pExist = FALSE;

  return nData;
}

int Script_GetStringSize(SCRIPT_FILE *pFile, const char *szName)
{
  SCRIPT_DATA *p;
  int          i, max;

  if(pFile && szName)
  {
    for(max = pFile->n_data, i = 0; i < max; i++)
    {
      p = pFile->data + i;
      if(script_compare_names(p->name, szName))
      {
        if(p->data_type == SDT_NULL)        return 1;
        else if(p->data_type == SDT_STRING) return p->data_size;
      }
    }
  }

  return -1;
}

int Script_GetNumberOfInts(SCRIPT_FILE *pFile, const char *szName)
{
  SCRIPT_DATA *p;
  int          i, max;

  if(pFile && szName)
  {
    for(max = pFile->n_data, i = 0; i < max; i++)
    {
      p = pFile->data + i;
      if(script_compare_names(p->name, szName))
      {
        if(p->data_type == SDT_NULL)        return 0;
        else if(p->data_type == SDT_NUMBER) return 1;
        else if(p->data_type == SDT_MULTI)  return p->data_size >> 2;
      }
    }
  }

  return -1;
}

BOOL Script_WriteString(SCRIPT_FILE *pFile, const char *szName, const char *szString, int nSize)
{
  char        *s, *name, *str;
  SCRIPT_DATA *p;
  int          i, l;

  if(!pFile || !szName) return FALSE;

  if     (!szString) { szString = script_null_str; nSize = 1; }
  else if(nSize <= 0)  nSize    = strlen(szString) + 1;

  p = (SCRIPT_DATA *)0;
  for(l = pFile->n_data, i = 0; i < l; i++)
  {
    if(script_compare_names(pFile->data[i].name, szName))
    { p = pFile->data + i; break; }
  }

  if(p)
  {
#ifdef SCRIPT_USE_TEMP_FILE
    long pos =
      (p->data_type == SDT_STRING || p->data_type == SDT_MULTI) && p->data_size >= nSize ? p->data_pos : -1;
    pos = script_tmp_write((STMPFILE *)pFile->tmp_file, pos, (void *)szString, nSize);
    if(pos == -1) return FALSE;
    p->data_pos  = pos;
    p->data_size = nSize;
    p->data_type = SDT_STRING;
#else
    if(p->data_type == SDT_STRING || p->data_type == SDT_MULTI)
    {
      if(p->data_size < nSize)
      {
        str = (char *)realloc(p->sz_data, nSize);
        if(!str) return FALSE;
        p->sz_data = str;
      }
    }
    else
    {
      str = (char *)malloc(nSize);
      if(!str) return FALSE;
      free(p->sz_data);
      p->sz_data = str;
    }

    p->data_type = SDT_STRING;
    memcpy(p->sz_data, szString, nSize);
    p->data_size = nSize;
#endif
  }
  else
  {
    if(pFile->n_alloc <= pFile->n_data) // must expand memory
    {
      p = (SCRIPT_DATA *)realloc(pFile->data, (pFile->n_alloc + 32) * sizeof(SCRIPT_DATA));
      if(!p) return FALSE;
      pFile->data    = p;
      pFile->n_alloc += 32;
    }

    p = pFile->data + pFile->n_data++;

    // get name size
    l = 1;
    s = (char *)szName;
    while((i = *(s++))) { if(!ISSPACE(i)) l++; }

    // alloc memory and copy name
    if(pFile->base_size < pFile->names_size + l)
    {
      name = pFile->names_base ? (char *)realloc(pFile->names_base, pFile->base_size + l + 1024) :
                                 (char *)malloc(pFile->base_size + l + 1024);
      if(!name) return FALSE;
      pFile->names_base    = name;
      pFile->base_size    += l + 1024;
      pFile->data[0].name = name;
      for(i = 1; i < pFile->n_data; i++)
        pFile->data[i].name = pFile->data[i - 1].name + pFile->data[i - 1].name_size;
    }
    else p->name = pFile->names_base + pFile->names_size;
    pFile->names_size += l;
    p->name_size       = l;
    str                = p->name;
    s                  = (char *)szName;
    while((i = *(s++)))
    {
      if(!ISSPACE(i))
      {
        if(INVALID_NAME_CHAR(i)) i = DEF_NAME_CHAR;
        *str++ = (char)i;
      }
    }
    *str = 0;

#ifdef SCRIPT_USE_TEMP_FILE
    {
      long pos = script_tmp_write((STMPFILE *)pFile->tmp_file, -1, (void *)szString, nSize);
      if(pos == -1) return FALSE;
      p->data_pos = pos;
    }
#else
    str = (char *)malloc(nSize);
    if(!str) { free(p); return FALSE; }
    memcpy(str, szString, nSize);
    p->sz_data   = str;
#endif
    p->data_size = nSize;
    p->data_type = SDT_STRING;
  }

  return TRUE;
}

BOOL Script_WriteInt(SCRIPT_FILE *pFile, const char *szName, int nValue)
{
  char        *ptr, *s;
  SCRIPT_DATA *p;
  int          i, max;
  long         l;

  if(!pFile || !szName) return FALSE;

  p = (SCRIPT_DATA *)0;
  for(max = pFile->n_data, i = 0; i < max; i++)
  {
    if(script_compare_names(pFile->data[i].name, szName))
    { p = pFile->data + i; break; }
  }

#ifdef SCRIPT_USE_TEMP_FILE
  if(!p)
#else
  if(p)
  {
    if(p->data_type == SDT_STRING) free(p->sz_data);
  }
  else
#endif
  {
    if(pFile->n_alloc <= pFile->n_data) // must expand memory
    {
      p = (SCRIPT_DATA *)realloc(pFile->data, (pFile->n_alloc + 32) * sizeof(SCRIPT_DATA));
      if(!p) return FALSE;
      pFile->data    = p;
      pFile->n_alloc += 32;
    }

    p = pFile->data + pFile->n_data++;

    // get name size
    l = 1;
    s = (char *)szName;
    while((i = *(s++))) { if(!ISSPACE(i)) l++; }


    // alloc memory and copy name
    if(pFile->base_size < pFile->names_size + l)
    {
      ptr = pFile->names_base ? (char *)realloc(pFile->names_base, pFile->base_size + l + 1024) :
                                (char *)malloc(pFile->base_size + l + 1024);
      if(!ptr) return FALSE;
      pFile->names_base    = ptr;
      pFile->base_size    += l + 1024;
      pFile->data[0].name = ptr;
      for(i = 1; i < pFile->n_data; i++)
        pFile->data[i].name = pFile->data[i - 1].name + pFile->data[i - 1].name_size;
    }
    else p->name = pFile->names_base + pFile->names_size;
    pFile->names_size += l;
    p->name_size       = l;
    ptr                = p->name;
    s                  = (char *)szName;
    while((i = *(s++)))
    {
      if(!ISSPACE(i))
      {
        if(INVALID_NAME_CHAR(i)) i = DEF_NAME_CHAR;
        *ptr++ = (char)i;
      }
    }
    *ptr = 0;
  }

  p->n_data    = nValue;
  p->data_size = 4;
  p->data_type = SDT_NUMBER;

  return TRUE;
}

BOOL Script_WriteBool(SCRIPT_FILE *pFile, const char *szName, BOOL fValue)
{
  char        *ptr, *s;
  SCRIPT_DATA *p;
  int          i, max;
  long         l;

  if(!pFile || !szName) return FALSE;

  p = (SCRIPT_DATA *)0;
  for(max = pFile->n_data, i = 0; i < max; i++)
  {
    if(script_compare_names(pFile->data[i].name, szName))
    { p = pFile->data + i; break; }
  }

#ifdef SCRIPT_USE_TEMP_FILE
  if(!p)
#else
  if(p)
  {
    if(p->data_type == SDT_STRING) free(p->sz_data);
  }
  else
#endif
  {
    if(pFile->n_alloc <= pFile->n_data) // must expand memory
    {
      p = (SCRIPT_DATA *)realloc(pFile->data, (pFile->n_alloc + 32) * sizeof(SCRIPT_DATA));
      if(!p) return FALSE;
      pFile->data    = p;
      pFile->n_alloc += 32;
    }

    p = pFile->data + pFile->n_data++;

    // get name size
    l = 1;
    s = (char *)szName;
    while((i = *(s++))) { if(!ISSPACE(i)) l++; }


    // alloc memory and copy name
    if(pFile->base_size < pFile->names_size + l)
    {
      ptr = pFile->names_base ? (char *)realloc(pFile->names_base, pFile->base_size + l + 1024) :
                                (char *)malloc(pFile->base_size + l + 1024);
      if(!ptr) return FALSE;
      pFile->names_base    = ptr;
      pFile->base_size    += l + 1024;
      pFile->data[0].name = ptr;
      for(i = 1; i < pFile->n_data; i++)
        pFile->data[i].name = pFile->data[i - 1].name + pFile->data[i - 1].name_size;
    }
    else p->name = pFile->names_base + pFile->names_size;
    pFile->names_size += l;
    p->name_size       = l;
    ptr                = p->name;
    s                  = (char *)szName;
    while((i = *(s++)))
    {
      if(!ISSPACE(i))
      {
        if(INVALID_NAME_CHAR(i)) i = DEF_NAME_CHAR;
        *ptr++ = (char)i;
      }
    }
    *ptr = 0;
  }

  p->n_data    = fValue ? TRUE : FALSE;
  p->data_size = 4;
  p->data_type = SDT_BOOL;

  return TRUE;
}

BOOL Script_WriteMultiInts(SCRIPT_FILE *pFile, const char *szName, const int *pnData, int nData)
{
  char        *s, *name, *str;
  SCRIPT_DATA *p;
  int          i, l, nSize;

  if(!pFile || !szName || !pnData || nData <= 0) return FALSE;

  nSize = nData * 4;

  p = (SCRIPT_DATA *)0;
  for(l = pFile->n_data, i = 0; i < l; i++)
  {
    if(script_compare_names(pFile->data[i].name, szName))
    { p = pFile->data + i; break; }
  }

  if(p)
  {
#ifdef SCRIPT_USE_TEMP_FILE
    long pos =
      (p->data_type == SDT_STRING || p->data_type == SDT_MULTI) && p->data_size >= nSize ? p->data_pos : -1L;
    pos = script_tmp_write((STMPFILE *)pFile->tmp_file, pos, (void *)pnData, nSize);
    if(pos == -1) return FALSE;
    p->data_pos  = pos;
    p->data_size = nSize;
    p->data_type = SDT_MULTI;
#else
    if(p->data_type == SDT_STRING || p->data_type == SDT_MULTI)
    {
      if(p->data_size < nSize)
      {
        str = (char *)realloc(p->sz_data, nSize);
        if(!str) return FALSE;
        p->sz_data = str;
      }
    }
    else
    {
      str = (char *)malloc(nSize);
      if(!str) return FALSE;
      free(p->sz_data);
      p->sz_data = str;
    }

    p->data_type = SDT_MULTI;
    memcpy(p->pn_data, pnData, nSize);
    p->data_size = nSize;
#endif
  }
  else
  {
    if(pFile->n_alloc <= pFile->n_data) // must expand memory
    {
      p = (SCRIPT_DATA *)realloc(pFile->data, (pFile->n_alloc + 32) * sizeof(SCRIPT_DATA));
      if(!p) return FALSE;
      pFile->data    = p;
      pFile->n_alloc += 32;
    }

    p = pFile->data + pFile->n_data++;

    // get name size
    l = 1;
    s = (char *)szName;
    while((i = *(s++))) { if(!ISSPACE(i)) l++; }

    // alloc memory and copy name
    if(pFile->base_size < pFile->names_size + l)
    {
      name = pFile->names_base ? (char *)realloc(pFile->names_base, pFile->base_size + l + 1024) :
                                 (char *)malloc(pFile->base_size + l + 1024);
      if(!name) return FALSE;
      pFile->names_base    = name;
      pFile->base_size    += l + 1024;
      pFile->data[0].name = name;
      for(i = 1; i < pFile->n_data; i++)
        pFile->data[i].name = pFile->data[i - 1].name + pFile->data[i - 1].name_size;
    }
    else p->name = pFile->names_base + pFile->names_size;
    pFile->names_size += l;
    p->name_size       = l;
    str                = p->name;
    s                  = (char *)szName;
    while((i = *(s++)))
    {
      if(!ISSPACE(i))
      {
        if(INVALID_NAME_CHAR(i)) i = DEF_NAME_CHAR;
        *str++ = (char)i;
      }
    }
    *str = 0;

#ifdef SCRIPT_USE_TEMP_FILE
    {
      long pos = script_tmp_write((STMPFILE *)pFile->tmp_file, -1, (void *)pnData, nSize);
      if(pos == -1) return FALSE;
      p->data_pos = pos;
    }
#else
    str = (char *)malloc(nSize);
    if(!str) { free(p); return FALSE; }
    memcpy(str, pnData, nSize);
    p->sz_data   = str;
#endif
    p->data_size = nSize;
    p->data_type = SDT_MULTI;
  }

  return TRUE;
}


#define gsputc(c, f) { if(putc(c, f) == EOF) return FALSE; }

BOOL Script_WriteStringDirect(FILE *pFile, const char *szName, const char *szString,
                              int nSize, const char *szComment, int nAlign)
{
  char *s;
  int   i, n, c, escape;

  if(!pFile || !szName) return FALSE;

  if     (!szString)  { szString = script_null_str; nSize = 1; }
  else if(nSize <= 0) { nSize = strlen(szString) + 1; }

  // writting comment
  if(szComment)
  {
    if(*szComment)
    {
      n = 0;
      s = (char *)szComment;
      while((c = *(s++))) { if(c == '\n') { n = 1; break; } }
      if(n) { if(fprintf(pFile, "\n/* %s */\n", szComment) < 0) return FALSE; }
      else  { if(fprintf(pFile, "\n// %s\n", szComment) < 0) return FALSE; }
    }
  }

  // writting name
  i = 0;
  s = (char *)szName;
  while((c = *(s++)))
  {
    if(!ISSPACE(c))
    {
      if(INVALID_NAME_CHAR(c)) c = DEF_NAME_CHAR;
      gsputc(c, pFile);
      i++;
    }
  }

  // align
  while(i++ < nAlign) { gsputc(' ', pFile); }

  gsputc(' ', pFile); gsputc('=', pFile); gsputc(' ', pFile);

  // writting the string
  s = (char *)szString;
  gsputc('\"', pFile);
  for(--nSize, i = 0; i < nSize; i++)
  {
    c = (int)(*s++);
    n = escape = 0;
    while(script_escape_eq[n] != -1)
    {
      if(script_escape_eq[n] == c) { escape = 1; break; }
      n++;
    }

    if(escape) // escape character
    {
      gsputc('\\', pFile);
      gsputc(script_escape_ch[n], pFile);
    }
    else
    {
      gsputc(c, pFile);
    }
  }
  gsputc('"', pFile); gsputc(';', pFile); gsputc('\n', pFile);

  return TRUE;
}

BOOL Script_WriteIntDirect(FILE *pFile, const char *szName, int nValue,
                           const char *szComment, int nAlign)
{
  char *s;
  int   c, i;
  enum {
    NO_COMMENT,
    LINE_COMMENT,
    COMMENT
  } n;

  if(!pFile || !szName) return FALSE;

  // verify comment
  if(szComment)
  {
    if(*szComment)
    {
      n = LINE_COMMENT;
      s = (char *)szComment;
      while((c = *(s++))) { if(c == '\n') { n = COMMENT; break; } }
    }
    else n = NO_COMMENT;
  }
  else n = NO_COMMENT;

  if(n == COMMENT) { if(fprintf(pFile, "/* %s */\n", szComment) < 0) return FALSE; }

  // writting name
  i = 0;
  s = (char *)szName;
  while((c = *(s++)))
  {
    if(!ISSPACE(c))
    {
      if(INVALID_NAME_CHAR(c)) c = DEF_NAME_CHAR;
      gsputc(c, pFile);
      i++;
    }
  }

  // align
  while(i++ < nAlign) { gsputc(' ', pFile); }

  // writting value
  if(n == LINE_COMMENT)
  {
    if((i = fprintf(pFile, " = %i;", nValue)) < 0) return FALSE;
    i = 16 - i;
    while(i-- > 0) { gsputc(' ', pFile); }
    if(fprintf(pFile, "// %s\n", szComment) < 0) return FALSE;
  }
  else
  { if(fprintf(pFile, " = %i;\n", nValue) < 0) return FALSE; }

  return TRUE;
}

BOOL Script_WriteBoolDirect(FILE *pFile, const char *szName, BOOL fValue,
                            const char *szComment, int nAlign)
{
  char *s;
  int   c, i;
  enum {
    NO_COMMENT,
    LINE_COMMENT,
    COMMENT
  } n;
  const char sz_TRUE[]  = "true";
  const char sz_FALSE[] = "false";

  if(!pFile || !szName) return FALSE;

  // verify comment
  if(szComment)
  {
    if(*szComment)
    {
      n = LINE_COMMENT;
      s = (char *)szComment;
      while((c = *(s++))) { if(c == '\n') { n = COMMENT; break; } }
    }
    else n = NO_COMMENT;
  }
  else n = NO_COMMENT;

  if(n == COMMENT) { if(fprintf(pFile, "/* %s */\n", szComment) < 0) return FALSE; }

  // writting name
  i = 0;
  s = (char *)szName;
  while((c = *(s++)))
  {
    if(!ISSPACE(c))
    {
      if(INVALID_NAME_CHAR(c)) c = DEF_NAME_CHAR;
      gsputc(c, pFile);
      i++;
    }
  }

  // align value
  while(i++ < nAlign) { gsputc(' ', pFile); }

  // writting value
  if(n == LINE_COMMENT)
  {
    if((i = fprintf(pFile, " = %s;", fValue ? sz_TRUE : sz_FALSE)) < 0) return FALSE;
    // align comment
    i = 16 - i;
    while(i-- > 0) { gsputc(' ', pFile); }
    if(fprintf(pFile, "// %s\n", szComment) < 0) return FALSE;
  }
  else
  { if(fprintf(pFile, " = %s;\n", fValue ? sz_TRUE : sz_FALSE) < 0) return FALSE; }

  return TRUE;
}

BOOL Script_WriteMultiIntsDirect(FILE *pFile, const char *szName, const int *pnData,
                                 int nData, const char *szComment, int nAlign)
{
  char       *s;
  int         i, n, c;
  const int  *nd;
  const char *format[3] = { "0x%.2X", "0x%.4X", "0x%.8X" };
  const int   nl[3]     = { 16, 12, 8 };

  if(!pFile || !szName || !pnData || nData <= 0) return FALSE;

  // writting comment
  if(szComment)
  {
    if(*szComment)
    {
      n = 0;
      s = (char *)szComment;
      while((c = *(s++))) { if(c == '\n') { n = 1; break; } }
      if(n) { if(fprintf(pFile, "\n/* %s */\n", szComment) < 0) return FALSE; }
      else  { if(fprintf(pFile, "\n// %s\n", szComment) < 0) return FALSE; }
    }
  }

  // writting name
  i = 0;
  s = (char *)szName;
  while((c = *(s++)))
  {
    if(!ISSPACE(c))
    {
      if(INVALID_NAME_CHAR(c)) c = DEF_NAME_CHAR;
      gsputc(c, pFile);
      i++;
    }
  }

  // align
  while(i++ < nAlign) { gsputc(' ', pFile); }
  gsputc(' ', pFile); gsputc('=', pFile); gsputc(' ', pFile);

  for(i = n = 0; i < nData; i++)
  { if((unsigned)pnData[i] > (unsigned)n) n = pnData[i]; }
  if     (n & 0xFF000000) c = 2;
  else if(n & 0xFF00)     c = 1;
  else                    c = 0;

  // writting data
  if(nData <= nl[c]) fprintf(pFile, "{ ");
  else               fprintf(pFile, "{\n  ");
  nd = pnData;
  for(i = 0; i < nData; i++)
  {
    if(i)
    {
      if((i % nl[c]) == 0) fprintf(pFile, ",\n  ");
      else                 fprintf(pFile, ", ");
    }
    fprintf(pFile, format[c], pnData[i]);
  }
  if(nData <= nl[c]) fprintf(pFile, " };\n");
  else               fprintf(pFile, "\n};\n");

  return TRUE;
}
