
#ifndef _VIDEO_H
#define _VIDEO_H

#ifdef SDL
#include "sdl.h"
#else
#include <pspdisplay.h>
#include <pspgu.h>
#endif
#include "def.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef struct _VIDEO_BUFFER {
  void *Data;
  int   Pitch;
  int   Width;
  int   Height;
#ifdef SDL
  void *SDLSurface;
  int   Locked;
#else
  int   Tw, Th;
#endif
} VIDEO_BUFFER;

#ifdef SDL

extern BOOL               sdl_init;
extern SDL_Surface       *main_surface;
extern SDL_Surface       *draw_buffer;
extern BOOL               draw_buffer_locked;
extern unsigned char      screen_buffer[256 * 192];
extern VIDEO_BUFFER       sdl_draw;
extern const VIDEO_BUFFER sdl_screen;

#else

#define SCREEN_BUFFER (unsigned char *)(0x44000000 + (512 * 272 * 4))

extern BOOL               gu_init;
extern int                gu_disp;
extern const VIDEO_BUFFER gu_draw[2][2];
extern const VIDEO_BUFFER gu_screen[2];

#endif

typedef enum _VIDEO_MODE {
  VM_NONE      = -1,
  VM_EMULATION = 0,
  VM_MENU      = 1
} VIDEO_MODE;

extern VIDEO_MODE video_mode;

#define Video_GetMode() (video_mode)

void          Video_Init(void);            // init video stuff
void          Video_Quit(void);            // quit video
void          Video_SetMode(VIDEO_MODE);   // set video mode: menu or emulation
void          Video_Clear(void);           // clear screen, draw and display buffers
VIDEO_BUFFER *Video_Alloc(int, int);       // alloc video memory
void          Video_Free(VIDEO_BUFFER *);  // free video memory
// display message in screen during emulation (valid only for VM_EMULATION video mode)
void          Video_ShowMessage(const char *, unsigned long);
void          Video_BlitEmulation(void);   // copy screen buffer to draw buffer
void          Video_Blit(VIDEO_BUFFER *, int, int, int, int, BOOL);
// load an image to memory
VIDEO_BUFFER *Video_LoadImage(const char *);

// default time to display in game messages
#define DEFAULT_MESSAGE_TIME 5000

#ifdef SDL

const VIDEO_BUFFER *Video_GetDrawBuffer(void);

#define Video_GetScreenBuffer() \
  (video_mode != VM_NONE ? (const VIDEO_BUFFER *)&sdl_screen : (const VIDEO_BUFFER *)0)
#define Video_Flip(vsync) {                                                                \
  if(sdl_init) {                                                                           \
    if(draw_buffer_locked) { SDL_UnlockSurface(draw_buffer); draw_buffer_locked = FALSE; } \
    SDL_BlitSurface(draw_buffer, NULL, main_surface, NULL);                                \
    SDL_UpdateRect(main_surface, 0, 0, draw_buffer->w, draw_buffer->h); } }


#else

#define Video_GetDrawBuffer() \
  (video_mode != VM_NONE ? (const VIDEO_BUFFER *)&gu_draw[(int)video_mode][gu_disp] : (const VIDEO_BUFFER *)0)
#define Video_GetScreenBuffer() \
  (video_mode != VM_NONE ? (const VIDEO_BUFFER *)&gu_screen[(int)video_mode] : (const VIDEO_BUFFER *)0)
#define Video_Flip(vsync) {                                                                \
  if(gu_init) {                                                                            \
    if(vsync) sceDisplayWaitVblankStart();                                                 \
    sceGuSwapBuffers();                                                                    \
    gu_disp ^= 1; } }

#endif

#ifdef __cplusplus
}
#endif

#endif // _VIDEO_H
